########## R script: EPandMFVBillustration ##########

# For illustration of expectation propagation
# and mean field variational Bayes for the 
# Normal random model used in Kim & Wand (2016).

# Last changed: 29 JAN 2016

# Set sample size and true parameter values:

n <- 100
muTrue <- 0 ; sigsqTrue <- 1

# Set the hyperparameters:

mu.mu <- 0 ; sigsq.mu <- 1e5 ; A <- 1e5

# Load required functions:

source("NorRSinferMFVB.r")
source("NorRSinferEP.r")
source("logScrA.r")        ;    source("eplogalpha.r")
source("logScrB.r")        ;    source("eplogbeta.r")
source("epg.r")            ;    source("epGN.r") 
source("epGIGone.r")       ;    source("epGIGtwo.r")
source("logmdigammaInv.r") ;    source("etaUpdate.r")
source("logPartN.r")       ;    source("logPartIG.r") 

# Generate data-set:

x <- rnorm(n,muTrue,sqrt(sigsqTrue))

# Obtain the expectation propagation approximations
# to the mean and variance prior density functions:

ansEP <- NorRSinferEP(x,mu.mu=mu.mu,sigsq.mu=sigsq.mu,A=A)
mu.q.mu.EP <- ansEP$mu.q.mu
sigsq.q.mu.EP <- ansEP$sigsq.q.mu
A.q.sigsq.EP <- ansEP$A.q.sigsq
B.q.sigsq.EP <- ansEP$B.q.sigsq
 
ansMFVB <- NorRSinferMFVB(x,mu.mu=mu.mu,sigsq.mu=sigsq.mu,A=A)
mu.q.mu.MFVB <- ansMFVB$mu.q.mu
sigsq.q.mu.MFVB <- ansMFVB$sigsq.q.mu
A.q.sigsq.MFVB <- ansMFVB$A.q.sigsq
B.q.sigsq.MFVB <- ansMFVB$B.q.sigsq


# Plot the approximate posterior density functions:

par(mfrow=c(2,2))

xg <- seq(mu.q.mu.EP - 3.5*sqrt(sigsq.q.mu.EP),
      mu.q.mu.EP + 3.5*sqrt(sigsq.q.mu.EP),length=1001)
plot(xg,dnorm(xg,mu.q.mu.EP,sqrt(sigsq.q.mu.EP)),col="darkgreen",
     bty="l",xlab=expression(mu),ylab="density",type="l",lwd=2)
mtext("expectation propagation approximate",line=3)
mtext(expression(paste("posterior density function for ",mu)),line=1)

xg <- seq(1/qgamma(shape=A.q.sigsq.EP,rate=B.q.sigsq.EP,0.999),
          1/qgamma(shape=A.q.sigsq.EP,rate=B.q.sigsq.EP,0.001),length=1001)
plot(xg,dgamma((1/xg),shape=A.q.sigsq.EP,rate=B.q.sigsq.EP)/(xg^2),col="darkgreen",
     bty="l",xlab=expression(sigma^2),ylab="density",type="l",lwd=2)
mtext("expectation propagation approximate",line=3)
mtext(expression(paste("posterior density function for ",sigma^2)),line=1)

xg <- seq(mu.q.mu.MFVB - 3.5*sqrt(sigsq.q.mu.MFVB),
      mu.q.mu.MFVB + 3.5*sqrt(sigsq.q.mu.MFVB),length=1001)
plot(xg,dnorm(xg,mu.q.mu.MFVB,sqrt(sigsq.q.mu.MFVB)),col="darkgreen",
     bty="l",xlab=expression(mu),ylab="density",type="l",lwd=2)
mtext("mean field variational Bayes approximate",line=3)
mtext(expression(paste("posterior density function for ",mu)),line=1)

xg <- seq(1/qgamma(shape=A.q.sigsq.MFVB,rate=B.q.sigsq.MFVB,0.999),
          1/qgamma(shape=A.q.sigsq.MFVB,rate=B.q.sigsq.MFVB,0.001),length=1001)
plot(xg,dgamma((1/xg),shape=A.q.sigsq.MFVB,rate=B.q.sigsq.MFVB)/(xg^2),col="darkgreen",
     bty="l",xlab=expression(sigma^2),ylab="density",type="l",lwd=2)
mtext("mean field variational Bayes approximate",line=3)
mtext(expression(paste("posterior density function for ",sigma^2)),line=1)

############ End of EPandMFVBillustration ############
