########## R function: NorRSinferEP ###########

# For performing approximate inference for the
# parameters in a Bayesian Normal random sample
# model via expectation propagation.

# Last changed: 14 AUG 2015

NorRSinferEP <- function(x,mu.mu=0,sigsq.mu=1e5,A=1e5,maxIter=100,
                         tolerance=1e-20,epsilon=0,plotML=FALSE,
                         convergeType="MaximumIterations")
{
   # Create function for approximate marginal log-likelihood:

   logML <- function(n,sumx,sumxsq,eta)
   {
      log.s.mu <- logPartN(eta$"p(mu)->mu" + eta$"p(x|mu,sigsq)->mu") - 0.5*log(2*pi)
  
      log.s.sigsq <- logPartIG(eta$"p(x|mu,sigsq)->sigsq" + eta$"p(sigsq|a)->sigsq")

      log.s.a <- logPartIG(eta$"p(a)->a" + eta$"p(sigsq|a)->a")

      log.s.pmu <- 0

      log.s.pxmusigsq <- (-0.5*(n+1)*log(2*pi) - lgamma(0.5*n-eta$"sigsq->p(x|mu,sigsq)"[1]-1)
               +eplogAlpha(0,eta$"mu->p(x|mu,sigsq)",eta$"sigsq->p(x|mu,sigsq)",c(n,sumx,sumxsq))$modulus
                            - logPartN(eta$"mu->p(x|mu,sigsq)" + eta$"p(x|mu,sigsq)->mu")
                            - logPartIG(eta$"sigsq->p(x|mu,sigsq)" + eta$"p(x|mu,sigsq)->sigsq"))
  
      log.s.psigsqa <-  (- 0.5*log(pi) -lgamma(-eta$"a->p(sigsq|a)"[1]-0.5)
                        + eplogBeta(0,-2,-eta$"sigsq->p(sigsq|a)"[2],
                                   0.5-eta$"sigsq->p(sigsq|a)"[1],
                                   eta$"a->p(sigsq|a)",
                                   c(0,eta$"sigsq->p(sigsq|a)"[2]),
                                   c(2,0,0))$modulus
                          - logPartIG(eta$"sigsq->p(sigsq|a)" +eta$"p(sigsq|a)->sigsq")
                          - logPartIG(eta$"a->p(sigsq|a)" + eta$"p(sigsq|a)->a"))
  

      log.s.pa <- 0

      ans <- log.s.mu + log.s.sigsq + log.s.a + log.s.pmu + log.s.pxmusigsq + log.s.psigsqa + log.s.pa

      return(ans)
   }
  
   # Extract sample size:

   n <- length(x)

   # Compute summary statistics:

   sumx <- sum(x) ; sumxsq <- sum(x^2)

   # Initialise the natural parameters
   # of the factor to node messages:

   eta <- vector("list",17)
   names(eta) <- c("p(mu)","p(a)","p(mu)->mu","mu->p(mu)","mu->p(x|mu,sigsq)",
                    "p(x|mu,sigsq)->mu", "p(x|mu,sigsq)->sigsq",
                    "sigsq->p(x|mu,sigsq)","sigsq->p(sigsq|a)",
                    "p(sigsq|a)->sigsq","p(sigsq|a)->a",
                    "a->p(sigsq|a)","a->p(a)","p(a)->a",
                    "q(mu)","q(sigsq)","q(a)")


   eta$"p(mu)" <- c(mu.mu/sigsq.mu,-0.5/sigsq.mu)
   eta$"p(a)" <- c(-1.5,-1/A^2)

   eta$"p(mu)->mu" <- c(mu.mu/sigsq.mu,-0.5/sigsq.mu)
   eta$"p(a)->a" <- c(-1.5,-1/A^2)
   eta$"p(x|mu,sigsq)->mu" <- c(0,-0.5)
   eta$"p(x|mu,sigsq)->sigsq" <- c(-2,-1)
   eta$"p(sigsq|a)->sigsq" <- c(-2,-1)
   eta$"p(sigsq|a)->a" <- c(-2,-1)

   # Update the message natural parameters
   # until convergence:

   converged <- FALSE ; itNum <- 0;
   logMLgrid <- rep(NA,maxIter)
   logMLcurr <- -Inf

   while (!converged)
   {
      logMLprev <- logMLcurr
      itNum <- itNum + 1
      
      # Update message to p(mu):
   
      eta$"mu->p(mu)" <- eta$"p(x|mu,sigsq)->mu"
   
      # Update messages to p(sigsq|a):
   
      eta$"sigsq->p(sigsq|a)" <- eta$"p(x|mu,sigsq)->sigsq"
      eta$"a->p(sigsq|a)" <- eta$"p(a)->a"
   
      # Update messages from p(sigsq|a):
 
      eta$"p(sigsq|a)->sigsq" <- (epsilon*eta$"p(sigsq|a)->sigsq"
                   +(1-epsilon)*etaUpdate(eta$"sigsq->p(sigsq|a)",
                                          eta$"a->p(sigsq|a)",
                                          aux=c(2,0,0),
                                          messageName="p(sigsq|a)->sigsq",
                                          engine="EP"))
   
      eta$"p(sigsq|a)->a" <- (epsilon*eta$"p(sigsq|a)->a"
                   +(1-epsilon)*etaUpdate(eta$"a->p(sigsq|a)",
                                          eta$"sigsq->p(sigsq|a)",
                                          aux=c(2,0,0),
                                          messageName="p(sigsq|a)->a",
                                          engine="EP"))
  
      # Update message to p(a):
   
      eta$"a->p(a)" <- eta$"p(sigsq|a)->a"
   
      # Update messages to p(x|mu,sigsq):
   
      eta$"mu->p(x|mu,sigsq)" <- eta$"p(mu)->mu"
      eta$"sigsq->p(x|mu,sigsq)" <- eta$"p(sigsq|a)->sigsq"
   
      # Update messages from p(x|mu,sigsq):
   
      eta$"p(x|mu,sigsq)->mu" <- (epsilon*eta$"p(x|mu,sigsq)->mu"
                   +(1-epsilon)*etaUpdate(eta$"mu->p(x|mu,sigsq)",
                                          eta$"sigsq->p(x|mu,sigsq)",
                                          aux=c(n,sumx,sumxsq),
                                          messageName="p(x|mu,sigsq)->mu",
                                          engine="EP"))
   
      eta$"p(x|mu,sigsq)->sigsq" <- (epsilon*eta$"p(x|mu,sigsq)->sigsq"
                   +(1-epsilon)*etaUpdate(eta$"sigsq->p(x|mu,sigsq)",
                                          eta$"mu->p(x|mu,sigsq)",
                                          aux=c(n,sumx,sumxsq),
                                          messageName="p(x|mu,sigsq)->sigsq",
                                          engine="EP"))

      if (itNum>=maxIter) 
      {
         converged <- TRUE
         #print("WARNING: maximum number of iterations exceeded.")
      }

      if (convergeType=="ApproxMargLogLik")
      {
         # Evaluate current approximate log-likelihood:

         logMLcurr <- logML(n,sumx,sumxsq,eta)
         logMLgrid[itNum] <- logMLcurr

         # Compute relative error:

         relErr <- abs((logMLcurr/logMLprev)-1)

         # Check `converged' conditions:

         if (relErr<tolerance) converged <- TRUE
      }
   }

   logMLgrid <- logMLgrid[1:itNum]
  
   useWaiting <- FALSE
   if (plotML)
   {
      omitVal <- 0
      iterGrid <- 1:length(logMLgrid)

      if (omitVal>0)
      {
         logMLgrid <- logMLgrid[-(1:omitVal)]
         iterGrid <- iterGrid[-(1:omitVal)]
      }
         
      plot(iterGrid,logMLgrid,type="b",bty="l",xlab="iterations",
           ylab="approximate marginal log-likelihood")
      ans <- readline()
      cat("Hit Enter to continue.\n")
   }
   
   # Now compute natural parameters of q densities:
   
   eta$"q(mu)"    <- eta$"p(mu)->mu" + eta$"p(x|mu,sigsq)->mu"
   eta$"q(sigsq)" <- eta$"p(x|mu,sigsq)->sigsq" + eta$"p(sigsq|a)->sigsq"
   eta$"q(a)" <-     eta$"p(sigsq|a)->a" + eta$"p(a)->a"
   
   # Convert natural parameters of q densities to 
   # common parameters:
   
   mu.q.mu <- -eta$"q(mu)"[1]/(2*eta$"q(mu)"[2])
   sigsq.q.mu <- -1/(2*eta$"q(mu)"[2])
   
   A.q.sigsq <- - eta$"q(sigsq)"[1] - 1
   B.q.sigsq <- - eta$"q(sigsq)"[2]

   # Return parameters of approximate posterior density functions:
   
   return(list(mu.q.mu=mu.q.mu,sigsq.q.mu=sigsq.q.mu,
               A.q.sigsq=A.q.sigsq,B.q.sigsq=B.q.sigsq))
}
   
########## End of NorRSinferEP ###########
   
